/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

#include <stdio.h>
#include <stdbool.h>
#include <pthread.h>

#define fence() __sync_synchronize()

// Le qualificatif `volatile` ci-dessous limite les optimisations
// que le compilateur C ferait en considérant qu'il s'agit d'un code
// séquentiel. Une alternative est de compiler avec -O0 seulement.
int volatile turn = 0;
bool volatile want[2] = { false, false };

void lock(int i) {
  int other = 1 - i;
  want[i] = true;
  turn = other;
  fence(); // nécessaire pour les architectures à mémoire faible
  while (want[other] && turn == other) ; // attente active

}

void unlock(int i) {
  want[i] = false;
}

int x = 0;

void *f(void *arg) {
  int n =  *((int *)arg);
  printf("starting %d\n", n);
  for (int i = 1; i <= 1000000; i++) {
    lock(n);
    x++;
    unlock(n);
  }
  pthread_exit(NULL);
}

int main(int argc, char **argv) {
  pthread_t t1;
  pthread_t t2;
  int c1 = 0;
  int c2 = 1;
  pthread_create(&t1, NULL, f, &c1);
  pthread_create(&t2, NULL, f, &c2);
  pthread_join(t1, NULL);
  pthread_join(t2, NULL);
  printf("x: %d\n",x);
}
